#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#define MAX(a,b) (((a)>(b))?(a):(b))

const float M_PI = static_cast<float>(4*atan(1.0));
float zNear = 0;
float fovy = 30.0;
float widthFr;
float depthFr;
float background[3], ambient[3], diffuse[3], specular[3], refractive[3], refractiveIndex;
unsigned int ambientMode, diffuseMode, specularMode, refractiveMode;
int shininess;
const unsigned int SETV = 0;
const unsigned int MULT = 1;
//if you have glut installed you can uncomment "#define USE_GLUT" and add the glut lib to linker
// under Visual Studio:  Properties | Linker | Input | Additional Dependencies : glut32.lib
//#define USE_GLUT

#ifdef USE_GLUT

#include <GL/glut.h>

enum Mousestatetype { ROTATE, TRANSLATE, ZOOM };
Mousestatetype mousestate;
int mouseX, mouseY;
float rotMatrix[16];
float transX = 0;
float transY = 0;
float transZ = 0;
#endif

float *geometry = 0;
float *color = 0;
unsigned int numPrimitives = 0;

float uS, uE, vS, vE, uStep, vStep;
float sphereRadius;
char outputFilename[256];

void clear()
{
	delete [] geometry;
	delete [] color;
}

void generate();

void readConfigFile(char *cfgFilename)
{
	FILE *file = fopen(cfgFilename,"r");
	fscanf(file,"%*s %*s %s\n",outputFilename);
	fscanf(file,"%*s %*s %f %*s %f %*s %f\n",&uS,&uE,&uStep);
	fscanf(file,"%*s %*s %f %*s %f %*s %f\n",&vS,&vE,&vStep);
	fscanf(file,"%*s %*s %f\n",&sphereRadius);
	
	fscanf(file,"%*s %f %f %f\n",&background[0],&background[1],&background[2]);
	char mode[5];
	fscanf(file,"%*s %s %f %f %f\n",mode,&ambient[0],&ambient[1],&ambient[2]);
	ambientMode = (strcmp(mode,"mult") == 0) ? MULT : SETV;
	fscanf(file,"%*s %s %f %f %f\n",mode,&diffuse[0],&diffuse[1],&diffuse[2]);
	diffuseMode = (strcmp(mode,"mult") == 0) ? MULT : SETV;
	fscanf(file,"%*s %s %f %f %f\n",mode,&specular[0],&specular[1],&specular[2]);
	specularMode = (strcmp(mode,"mult") == 0) ? MULT : SETV;
	fscanf(file,"%*s %s %f %f %f\n",mode,&refractive[0],&refractive[1],&refractive[2]);
	refractiveMode = (strcmp(mode,"mult") == 0) ? MULT : SETV;
	fscanf(file,"%*s %f\n",&refractiveIndex);
	fscanf(file,"%*s %d\n",&shininess);
	widthFr = MAX(uE-uS,vE-vS)+2*sphereRadius;
	depthFr = 3*(vE-vS+2*sphereRadius);
	zNear = widthFr/(2*tan(fovy/360.0f*M_PI));
}

#ifdef USE_GLUT
void init(char *filename) 
{
	glClearColor (0, 0, 0, 0);
	glShadeModel (GL_FLAT);
	glEnable(GL_DEPTH_TEST);
	for(int i = 0; i < 4; i++) {
		for(int j = 0; j < i; j++) {
			rotMatrix[i*2+j] = 0.0;
		}
	}
	rotMatrix[0] = rotMatrix[5] = rotMatrix[10] = rotMatrix[15] = 1.0;

	glPushMatrix();
	glLoadIdentity();
	glRotatef(-90,1.0, 0.0, 0.0);
	glMultMatrixf(rotMatrix);
	glGetFloatv(GL_MODELVIEW_MATRIX,rotMatrix);
	glPopMatrix();


	readConfigFile(filename);
	transZ = -depthFr/2.0-zNear;

	generate();
}

void display(void)
{
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(transX,transY,transZ);
	glMultMatrixf(rotMatrix);

	glColorPointer(3,GL_FLOAT,0,color);
	glVertexPointer(3,GL_FLOAT,0,geometry);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glDrawArrays(GL_POINTS,0,numPrimitives);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);


	glPopMatrix();
	glutSwapBuffers();
	glutPostRedisplay();
}

void reshape (int w, int h)
{
	glViewport (0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	//gluPerspective(30,w/h,1,1+1.1*(vE-vS));
	float wh = widthFr/2;
	glFrustum(-wh,wh,-wh,wh,zNear,zNear+depthFr);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void mousemotion (int x, int y)
{
	int dx = mouseX - x;
	int dy = mouseY - y;
	switch (mousestate) {
	case ROTATE:
		glPushMatrix();
		glLoadIdentity();
		glRotatef(-0.5*dx,0.0, 1.0, 0.0);
		glRotatef(-0.5*dy,1.0, 0.0, 0.0);
		glMultMatrixf(rotMatrix);
		glGetFloatv(GL_MODELVIEW_MATRIX,rotMatrix);
		glPopMatrix();
		break;
	case TRANSLATE:
		transX -= 0.005*dx;
		transY += 0.005*dy;
		break;
	case ZOOM:
		transZ += 0.01*dy;
		break;
	default:
		break;
	}
	mouseX = x;
	mouseY = y;

	glutPostRedisplay();
}

void mouse (int button, int state, int x, int y)
{
	if (state == GLUT_DOWN) {
		mouseX = x;
		mouseY = y;
		switch (button) {
		case GLUT_LEFT_BUTTON:
			mousestate = ROTATE;
			break;
		case GLUT_MIDDLE_BUTTON:
			mousestate = TRANSLATE;
			break;
		case GLUT_RIGHT_BUTTON:
			mousestate = ZOOM;
			break;
		default:
			break;
		}
	}
}

void keyboard (unsigned char key, int x, int y)
{
	switch (key) {
		case 27:
			clear();
			exit (0);
			break;
	}
}
#endif

void HSV2RGB(const float *hsv, float *rgb)
{
	/* this Code is copied from Computer Graphics, Principles an Practice, Foley et al */
  /* Given: h in [0,360] or UNDEFINED, s and v in [0,1] */
	/* Desired: r,g,b, each in [0,1] */
	float h = hsv[0];
	float s = hsv[1];
	float v = hsv[2];
	float &r = rgb[0];
	float &g = rgb[1];
	float &b = rgb[2];
	if(s==0) {
		if(h==-1.0) { // UNDEFINED:=-1.0
			r = v;
			g = v;
			b = v;
		} else {
			fprintf(stderr,"hsv -> rgb conversion error (s==0 and h > 0)\n");
			exit(1);
		}
	} else {
		float f,p,q,t;
		int i;

		if (h==360.0)
			h = 0.0;
		h /= 60.0;
		i = static_cast<int>(floor(h));
		f = h-i;
		p = v*(1.0f-s);
		q = v*(1.0f-(s*f));
		t = v*(1.0f-(s*(1.0f-f)));
		switch(i) {
				case 0: r = v; g = t; b = p; break;
				case 1: r = q; g = v; b = p; break;
				case 2: r = p; g = v; b = t; break;
				case 3: r = p; g = q; b = v; break;
				case 4: r = t; g = p; b = v; break;
				case 5: r = v; g = p; b = q; break;
		}
	}
}


FILE* openXML(char *filename)
{
	FILE *file = fopen(filename,"w");
	fprintf(file,"<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n");
	fprintf(file,"<!--  Steinbach Screw\n");
	fprintf(file," -->\n");
	fprintf(file,"<scene>\n");
	//fprintf(file,"  <camera pos=\"0 7 0\" dir=\"0 -1 0\" up=\"0 0 1\" focallength=\"1\" window=\"-1 -1 1 1\"/>\n");
	float wh = widthFr/2.0f;
	fprintf(file,"  <camera pos=\"0 %f 0\" dir=\"0 1 0\" up=\"0 0 1\" focallength=\"%f\" window=\"%f %f %f %f\"/>\n",
		-zNear-depthFr/2.0f,zNear,-wh,-wh,wh,wh);

	return file;
}

void openMaterialXML(FILE *file)
{
	fprintf(file,"  <materials>\n");
	fprintf(file,"    <background color=\"%f %f %f\"/>\n",background[0],background[1],background[2]);
}

void closeMaterialXML(FILE *file)
{
	fprintf(file,"  </materials>\n");
}

void writeMaterialXML(FILE *file, float *col, int id)
{
	static float mat[12];
	if(ambientMode == MULT) 
		for(int i = 0; i < 3; i++) mat[i] = ambient[i]*col[i];
	else
		for(int i = 0; i < 3; i++) mat[i] = ambient[i];

	if(diffuseMode == MULT) 
		for(int i = 0; i < 3; i++) mat[i+9] = diffuse[i]*col[i];
	else
		for(int i = 0; i < 3; i++) mat[i+9] = diffuse[i];

	if(specularMode == MULT) 
		for(int i = 0; i < 3; i++) mat[i+3] = specular[i]*col[i];
	else
		for(int i = 0; i < 3; i++) mat[i+3] = specular[i];

	if(refractiveMode == MULT) 
		for(int i = 0; i < 3; i++) mat[i+6] = refractive[i]*col[i];
	else
		for(int i = 0; i < 3; i++) mat[i+6] = refractive[i];
	
	fprintf(file,"   <material type=\"phong\" id=\"%d\" ambient=\"%f %f %f\" specular=\"%f %f %f\" refractive=\"%f %f %f\" refractiveIndex=\"%f\" diffuse=\"%f %f %f\" shininess=\"%d\"/>\n",
		id,mat[0],mat[1],mat[2],mat[3],mat[4],mat[5],mat[6],mat[7],mat[8],refractiveIndex,mat[9],mat[10],mat[11],shininess);
}

void openGeometryXML(FILE *file)
{
	fprintf(file,"  <geometry>\n");
}

void closeGeometryXML(FILE *file)
{
	fprintf(file,"  </geometry>\n");
}

void writeSphereXML(FILE *file, float *center, float radius, int materialID)
{
	fprintf(file,"    <surface type=\"sphere\">\n");
	fprintf(file,"        <sphere center=\"%f %f %f\" radius=\"%f\" materialID=\"%d\"/>\n",center[0],center[1],center[2],radius,materialID);
	fprintf(file,"    </surface>\n");
}
inline void getSteinbach(float *x, float u, float v)
{
	x[0] = u * cos (v);
	x[1] = u * sin (v);
	x[2] = v * cos (u);
}

void writeDirectionalLightXML(FILE *file)
{
	float zPos = zNear+depthFr/2.0f;
	float dev = zPos*sin(1.0f/18.0f*M_PI);
	fprintf(file,"  <lights>\n");
	fprintf(file,"    <light type=\"directionalLight\" direction=\"%f %f %f\" color=\"1.0 1.0 1.0\"/>\n",-dev,-zPos,-dev);
	fprintf(file,"  </lights>\n");
}

void writePointLightXML(FILE *file)
{
	fprintf(file,"  <lights>\n");
	fprintf(file,"    <light type=\"pointLight\" position=\"0 0 0\" color=\"1.0 1.0 1.0\"/>\n");
	fprintf(file,"  </lights>\n");
}

void closeXML(FILE *file)
{
	fprintf(file,"</scene>\n");
	fclose(file);
}

// generate XML file of steinbach screw
void generate()
{
	const unsigned int ures = static_cast<unsigned int>(ceil((uE-uS)/uStep));
	const unsigned int vres = static_cast<unsigned int>(ceil((vE-vS)/vStep));
	numPrimitives = ures*vres;
	float Du = uE-uS;
	float Dv = vE-vS;
	const float paramscale = sqrt(Du*Du+Dv*Dv);
	geometry = new float[3*ures*vres];
	color = new float[3*ures*vres];
	FILE *file = openXML(outputFilename);
	float hsv[3] = {0, 1, 1};
	float u = uS, v;
	float du,dv;

	// generate geometry and color arrays, also used as vertex arrays for visualization in OpenGL
	for(unsigned int uc = 0; uc < ures; uc++)
	{
		v = vS;
		du = u-uS;
		for(unsigned int vc = 0; vc < vres; vc++) 
		{
			dv = v-vS;
			unsigned int idx = 3*(vres*uc+vc);
			getSteinbach(&geometry[idx],u,v);
			hsv[0] = 360.0f*sqrt(static_cast<float>(du*du+dv*dv))/paramscale; 
			HSV2RGB(hsv,&color[idx]);
			v += vStep;
		}
		u += uStep;
	}

	// write materials to XML-file
	openMaterialXML(file);
	for(unsigned int i = 0; i < ures*vres; i++)
	{
		writeMaterialXML(file,&color[3*i],i);
	}
	closeMaterialXML(file);

	// write geometry (spheres) to XML-file
	openGeometryXML(file);
	for(unsigned int i = 0; i < ures*vres; i++)
	{
		writeSphereXML(file,&geometry[3*i],sphereRadius,i);
	}
	closeGeometryXML(file);

	//writePointLightXML(file);
	writeDirectionalLightXML(file);
	closeXML(file);

	return ;
}

// INUT ARGUMENT: config file which specifies the steinbach params, sphere radius and name of the XML-file for the raytracer
int main(int argc, char **argv)
{
	if(argc < 2) {
		fprintf(stderr,"wrong argument list!\n");
		getchar();
		return 1;
	}
// if USE_GLUT is set, the steinbach screw is visualized as point cloud
#ifdef USE_GLUT
	glutInit(&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize (300, 300); 
	glutInitWindowPosition (100, 100);
	glutCreateWindow (argv[0]);
	init(argv[1]);


	glutDisplayFunc(display); 
	glutMouseFunc(mouse);
	glutMotionFunc(mousemotion);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);

	glutMainLoop();
// otherwise its only written to file
#else
	readConfigFile(argv[1]);
	generate();
#endif
	clear();
	return 0;
}
